/*  FreeJ
 *  (c) Copyright 2001 Denis Roio aka jaromil <jaromil@dyne.org>
 *
 * This source code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Public License as published 
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * Please refer to the GNU Public License for more details.
 *
 * You should have received a copy of the GNU Public License along with
 * this source code; if not, write to:
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <iostream>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#include <v4l.h>

/* CHANGE THIS if you want to grab from another channel on your card
   (i.e. svideo, TVin ... ) */
#define INPUT_CHANNEL 1

V4lGrabber::V4lGrabber() {
  dev = -1;
  input = INPUT_CHANNEL;
  buffer = NULL;
  
  synchronous.registerObjectName("V4lGrabber");
}

V4lGrabber::~V4lGrabber() {
  _close();
}

void V4lGrabber::_close() {
  notice("Closing video4linux grabber layer");

  if(device!=NULL) {
    act("closing device %s :: %u", device,dev);
    free(device);
  }

  if(dev>0)
    close(dev);
  
  if(buffer!=NULL) {
    act("unmapping address 0x%lx sized %u bytes",(long unsigned int)buffer,grab_map.size);
    munmap(buffer,grab_map.size);
  }

  jfree(rgb_surface);
}

bool V4lGrabber::detect(char *devfile) {
  int counter, res;
  char *capabilities[] = {
    "VID_TYPE_CAPTURE          can capture to memory",
    "VID_TYPE_TUNER            has a tuner of some form",
    "VID_TYPE_TELETEXT         has teletext capability",
    "VID_TYPE_OVERLAY          can overlay its image onto the frame buffer",
    "VID_TYPE_CHROMAKEY        overlay is chromakeyed",
    "VID_TYPE_CLIPPING         overlay clipping supported",
    "VID_TYPE_FRAMERAM         overlay overwrites frame buffer memory",
    "VID_TYPE_SCALES           supports image scaling",
    "VID_TYPE_MONOCHROME       image capture is grey scale only",
    "VID_TYPE_SUBCAPTURE       capture can be of only part of the image"
  };

  func("V4lGrabber::detect()");

  device = strdup(devfile);

  if (-1 == (dev = open(device,O_RDWR))) {
    error("error in opening video capture device");
    return(false);
  }
  
  res = ioctl(dev,VIDIOCGCAP,&grab_cap);
  if(res<0) {
    error("error in VIDIOCGCAP ");
    return(false);
  }

  if(get_debug()>0) {
    
    notice("Device detected is %s",device);
    act("%s",grab_cap.name);
    act("%u channels detected",grab_cap.channels);
    act("max size w[%u] h[%u] - min size w[%u] h[%u]",grab_cap.maxwidth,grab_cap.maxheight,grab_cap.minwidth,grab_cap.minheight);
    act("Video capabilities:");
    for (counter=0;counter<11;counter++)
      if (grab_cap.type & (1 << counter)) act("%s",capabilities[counter]);

    if (-1 == ioctl(dev, VIDIOCGPICT, &grab_pic)) {
      perror("ioctl VIDIOCGPICT ");
      exit(1);
    }
    
    if (grab_pic.palette & VIDEO_PALETTE_GREY)
      act("VIDEO_PALETTE_GREY        device is able to grab greyscale frames");
  }    
  
  if(strncmp(grab_cap.name,"Buz",3)==0)
    /* the device is a iomega Buz. We need to check that 'cause the iomegaBuz
       claims to have the VIDEO_PALETTE_GREY capability, but then grabs 24bit
       images, overflowing the buffer. So we don't trust the capability report
       and turn on software 24bit->8bit downsampling */
    grab24 = 1;
  
  if(strncmp(grab_cap.name,"OV511",5)==0)
    /* the device is a USB camera, chipset OV511, wich has only one input
       channel (afaik, as long as i do not have one) so we force that chan */
    input = 0;
  
  if(grab_cap.type & VID_TYPE_TUNER)
    /* the device does'nt has any tuner, so we avoid some ioctl
       this should be a fix for Creative Webcam II. thanks to Ben Wilson */
    have_tuner = 1;
  
  /* set the minwidth and minheight */
  minw = grab_cap.minwidth;
  minh = grab_cap.minheight;
  
  if (ioctl (dev, VIDIOCGMBUF, &grab_map) == -1) {
    error("error in ioctl VIDIOCGMBUF");
    return(false);
  }
  /* print memory info */
  if(get_debug()>0) {
    act("memory map of %i frames: %i bytes",grab_map.frames,grab_map.size);
    for(counter=0;counter<grab_map.frames;counter++)
      act("Offset of frame %i: %i",counter,grab_map.offsets[counter]);
  }
  num_frame = grab_map.frames;
  return(true);
}

/* v4l palettes 
   VIDEO_PALETTE_GREY     b/w
   VIDEO_PALETTE_HI240    8bit
   VIDEO_PALETTE_RGB565   16bit
   VIDEO_PALETTE_RGB24    24bit
   VIDEO_PALETTE_YUV422   YUV8bit (4Y 2U 2V)
   others... */

bool V4lGrabber::init(Context *screen,int wdt, int hgt) {
  int i;
  func("V4lGrabber::init()");

  /* set image source and TV norm */
  grab_chan.channel = input;
  
  if(have_tuner) {
    /* does this only if the device has a tuner */
    if (-1 == ioctl(dev,VIDIOCGCHAN,&grab_chan)) {
      error("error in ioctl VIDIOCGCHAN ");
      return(false);
    }
    
    if (-1 == ioctl(dev,VIDIOCSCHAN,&grab_chan)) {
      error("error in ioctl VIDIOCSCHAN ");
      return(false);
    }
  }
  
  if (-1 == ioctl(dev, VIDIOCGPICT, &grab_pic)) {
    error("error in ioctl VIDIOCGPICT ");
    return(false);
  }

  /* TODO: check with minwidth maxwidth */

  w = (wdt == 0) ? screen->w : wdt;
  h = (hgt == 0) ? screen->h : hgt;
  bpp = screen->bpp;
  size = w*h*(bpp>>3);
  pitch = w*(bpp>>3);
  fps = &screen->fps;

  palette = VIDEO_PALETTE_YUV422P;
  /* choose best yuv2rgb routine (detecting cpu) */
  yuv2rgb = yuv2rgb_init(bpp,0x1); /* arg2 is MODE_RGB */
  rgb_surface = jalloc(rgb_surface,size);
  /* best until now
     u = (w*h);
     v = u+(u/2);
  */
     u = (w*h);
     v = u+(u/2);
  
  for(i=0; i<grab_map.frames; i++) {
    grab_buf[i].format = palette;
    grab_buf[i].frame  = i;
    grab_buf[i].height = h;
    grab_buf[i].width = w;
  }
  
  cur_frame = ok_frame = 0;
  
  /* mmap (POSIX.4) buffer for grabber device */
  buffer = (unsigned char *) mmap(0,grab_map.size,PROT_READ|PROT_WRITE,MAP_SHARED,dev,0);
  if(MAP_FAILED == buffer) {
    error("cannot allocate v4lgrabber buffer ");
    return(false);
  }
  
  /* grab the first frames */
  for(i=0;i<grab_map.frames;i++) {
    if (-1 == ioctl(dev,VIDIOCMCAPTURE,&grab_buf[i])) {
      error("error in ioctl VIDIOCMCAPTURE");
      return(false);
    }
  }

  feeded = true;

  /* INIT from the LAYER CLASS */
  _init(screen);

  notice("V4L layer :: w[%u] h[%u] bpp[%u] size[%u] grab_mmap[%u]",w,h,bpp,size,size*num_frame);
  
  return(true);
}

void *V4lGrabber::get_buffer() {
  /*  return(&buffer[grab_map.offsets[ok_frame]]); */
  (*yuv2rgb)((uint8_t *) rgb_surface,
	     (uint8_t *) &buffer[grab_map.offsets[ok_frame]],
	     (uint8_t *) &buffer[grab_map.offsets[ok_frame]+u],
	     (uint8_t *) &buffer[grab_map.offsets[ok_frame]+v],
	     w, h, pitch, w, w);
  return(rgb_surface);
}

bool V4lGrabber::feed() {
  ok_frame = cur_frame;
  cur_frame = ((cur_frame+1)%num_frame);
  grab_buf[0].format = palette;
  
  if (-1 == ioctl(dev,VIDIOCSYNC,&grab_buf[cur_frame])) {
    func("V4lGrabber::feed");
    error("error in ioctl VIDIOCSYNC");
  }

  if (-1 == ioctl(dev,VIDIOCMCAPTURE,&grab_buf[cur_frame])) {
    func("V4lGrabber::feed");
    error("error in ioctl VIDIOCMCAPTURE");
  }
  
  feeded = true;
  return(feeded);
  
}
